#include <linux/module.h>
#include <linux/moduleparam.h>
#include <asm/uaccess.h>
#include <linux/parport.h>
#include <linux/miscdevice.h>
#include <linux/types.h>
#include <linux/init.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,18)
#error "You need at least Linux 2.4.18 for this driver"
#endif

/*#define debug*/
#include "debug.h"

MODULE_AUTHOR("Dirk Jagdmann <doj@cubic.org");
MODULE_DESCRIPTION("driver for LED devices connected to parallel port");
MODULE_SUPPORTED_DEVICE("led");
MODULE_LICENSE("GPL");

int parport=0;
MODULE_PARM(parport, "1-4i");
MODULE_PARM_DESC(parport, "parport to use for output");

int updateall=0;
MODULE_PARM(updateall, "i");
MODULE_PARM_DESC(updateall, "update all 4 segments of LED32 hardware");

#define MODULE_NAME "LED"

#define true 1
#define false 0

#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2

struct parport* PARport=0;

/* parallel port defs */
#define STROBE0 PARPORT_CONTROL_STROBE
#define STROBE1 0
#define AUTOFEED0 PARPORT_CONTROL_AUTOFD
#define AUTOFEED1 0
#define INIT0 0
#define INIT1 PARPORT_CONTROL_INIT
#define SELECT0 PARPORT_CONTROL_SELECT
#define SELECT1 0

#define LEDE0 SELECT0
#define LEDE1 SELECT1
#define LEDA0 AUTOFEED0
#define LEDA1 AUTOFEED1
#define LEDB0 INIT0
#define LEDB1 INIT1

#define LEDMINORNUMBER 151

/** kernel function called when device is opened */
int FILEopen(struct inode *inode, struct file *file)
{
  PRINTK(KERN_INFO MODULE_NAME ": open()\n");
  return 0;
}

enum { LED8, LED12, LED32 };

int ledtype=LED8;
MODULE_PARM(ledtype, "i");
MODULE_PARM_DESC(ledtype, "type of leds connected: 0: 8 leds, 1: 12 leds, 2: 32 leds");

void LEDout8(uint8_t data)
{
  PARport->ops->write_data(PARport, data);
}

void LEDout12(uint32_t data)
{
  uint8_t command=0;
  if(data&0x001)
    command|=STROBE1;
  else
    command|=STROBE0;
  if(data&0x200)
    command|=AUTOFEED1;
  else
    command|=AUTOFEED0;
  if(data&0x400)
    command|=INIT1;
  else
    command|=INIT0;
  if(data&0x800)
    command|=SELECT1;
  else
    command|=SELECT0;
  PARport->ops->write_data(PARport, (data>>1)&0xff);
  PARport->ops->write_control(PARport, command);
}

void LEDout32(uint8_t data, int num)
{
  /** 2 bits to select latch */
  const unsigned char LEDcommand[]={LEDB0|LEDA0, LEDB0|LEDA1, LEDB1|LEDA0, LEDB1|LEDA1};
  const unsigned char command=LEDcommand[num];

  PARport->ops->write_control(PARport, command|LEDE1);
  PARport->ops->write_data(PARport, data);
  PARport->ops->write_control(PARport, command|LEDE0);
  PARport->ops->write_control(PARport, command|LEDE1);
}

/** clear the LED bar */
void LEDclear(void)
{
  int i;
  switch(ledtype)
    {
    default:
    case LED8:
      LEDout8(0);
      PARport->ops->write_control(PARport, STROBE0|AUTOFEED0|INIT0|SELECT0);
      break;
    case LED12:
      LEDout12(0);
      break;
    case LED32:
      for(i=0; i<4; i++)
	LEDout32(0, i);
      break;
    }
}

/** kernel function called when data is written to the LED. */
ssize_t LEDFILEwrite(struct file *file, const char *buf, size_t size, loff_t *pos)
{
  uint32_t new;
  int i;

  PRINTK(KERN_INFO MODULE_NAME ": write()\n");

  if (verify_area (VERIFY_READ, buf, size)) 
    return -EINVAL;

  copy_from_user (&new, buf, sizeof(uint32_t));

  switch(ledtype)
    {
    default:
    case LED8:
      {
	// we have to condense 32 bits to 8 bits
	static const int tab[] = {
	  0,0,0,0,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,4,5,5,5,5,6,6,6,6,7,7,7,7
	};
	uint8_t n=0;
	for(i=0; i<32; i++)
	  if(new&(1<<i))
	    n|=(1<<tab[i]);
	LEDout8(n);
      }
      break;
    case LED12:
      {
	// we have to condense 32 bits to 12 bits
	static const int tab[] = {
	  0,0,1,1,1,2,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,8,8,8,9,9,10,10,11,11
	};
	uint32_t n=0;
	for(i=0; i<32; i++)
	  if(new&(1<<i))
	    n|=(1<<tab[i]);
	LEDout12(n);
      }
      break;
    case LED32:
      {
	static uint32_t old=0;
	const uint32_t new2=new;
	for(i=0; i<sizeof(uint32_t); i++)
	  {
	    register uint8_t n=new&0xff;
	    register uint8_t o=old&0xff;
	  
	    if(updateall || n!=o)
	      LEDout32(n, i);
	  
	    new>>=8;
	    old>>=8;
	  }
	old=new2;
      }
      break;
    }

  file->f_pos += sizeof(uint32_t);
  return sizeof(uint32_t);
}

/** kernel function called when device is closed */
int LEDFILEclose(struct inode *inode, struct file *file)
{
  PRINTK(KERN_INFO MODULE_NAME ": close()\n");
  LEDclear();
  return 0;
}

/** kernel structure used to register device operations on LED */
struct file_operations LEDFILEops =
{
  .write = LEDFILEwrite,
  .open = FILEopen,
  .release = LEDFILEclose,
};

/** kernel structure used to register LED device */
struct miscdevice LEDmiscDevice =
{
  .minor = LEDMINORNUMBER,
  .name = "led",
  .fops = &LEDFILEops
};



/*****************************************************************************/
/* Parport Functions */
/*****************************************************************************/

struct pardevice *PARdevice=0;
struct parport* PARports[PARPORT_MAX];

#if 0
int PARpreempt(void *data)
{
#if 0
  if (busy_right_now)
    return 1;

  must_reclaim_port = 1;
#endif 
  return 0;
}

void PARwakeup(void *data)
{
#if 0
  struct toaster *private = handle;
  struct pardevice *dev = private->dev;
  if (!dev) return; /* avoid races */

  if (want_port)
    parport_claim (dev);
#endif
}
#endif

void PARattach(struct parport *port)
{
  PRINTK(KERN_INFO MODULE_NAME ": PARattach: %i %s\n", port->number, port->name);
  if(port->number<PARPORT_MAX)
    PARports[port->number]=port;
}

void PARdetach(struct parport *port)
{
  PRINTK(KERN_INFO MODULE_NAME ": PARdetach: %i %s\n", port->number, port->name);
}

struct parport_driver PARdriver = {
  .name = "LED driver",
  .attach = PARattach, 
  .detach = PARdetach, 
};

/*****************************************************************************/
/* Module Functions */
/*****************************************************************************/

void __exit exit_lcd(void)
{
  if(PARport)
    {
      LEDclear();
      if(misc_deregister(&LEDmiscDevice) < 0)
	printk(KERN_INFO MODULE_NAME ": unregister LED minor %i failed\n", LEDMINORNUMBER);

      PARport->ops->disable_irq(PARport);
      parport_release(PARdevice);
      parport_unregister_device(PARdevice);
    }
  
  parport_unregister_driver(&PARdriver);

  printk (KERN_INFO MODULE_NAME ": uninstalled\n");
}

int __init init_lcd(void)
{
  int i;
  void *data=0;

  PRINTK(KERN_INFO MODULE_NAME ": init trying parport %i\n", parport);

  /* initialize parport stuff */
  for(i=0; i<PARPORT_MAX; i++)
    PARports[i]=0;

  if(parport_register_driver(&PARdriver))
    {
      printk(KERN_INFO MODULE_NAME ": register parport driver failed\n");
      return -EIO;
    }

  if(!PARports[parport])
    {
      printk(KERN_INFO MODULE_NAME ": parport %i not found\n", parport);
      exit_lcd();
      return -EIO;
    }
  PARport=PARports[parport];

  PRINTK(KERN_INFO MODULE_NAME ": I am registering port %i\n", parport);
  PARdevice=parport_register_device(PARport, 
				    MODULE_NAME, 
				    NULL, /*PARpreempt*/
				    NULL, /*PARwakeup*/
				    NULL,
				    0 /*flags*/, 
				    data);

  if(PARdevice==0)
    {
      printk(KERN_INFO MODULE_NAME ": could not register port %i\n", parport);
      exit_lcd();
      return -EIO;
    }

  PRINTK(KERN_INFO MODULE_NAME ": I am claiming port %i\n", parport);
  i=parport_claim_or_block(PARdevice);
  if(i<0)
    {
      printk(KERN_INFO MODULE_NAME ": could not claim port %i\n", parport);
      exit_lcd();
      return -EIO;
    }
  PRINTK(KERN_INFO MODULE_NAME ": successfully claimed port %i\n", parport);

  if(ledtype==8)
    ledtype=LED8;
  else if(ledtype==12)
    ledtype=LED12;
  else if(ledtype==32)
    ledtype=LED32;
  if(ledtype>LED32)
    {
      printk (KERN_ERR MODULE_NAME ": wron ledtype parameter\n");
      exit_lcd();
      return -EIO;
    }
  printk(KERN_INFO MODULE_NAME ": using ledtype %i\n", ledtype);

  LEDFILEops.owner = THIS_MODULE;
  if (misc_register(&LEDmiscDevice) < 0)
    {
      printk (KERN_ERR MODULE_NAME ": register of LED minor %i failed\n", LEDMINORNUMBER);
      exit_lcd();
      return -EIO;
    }

  LEDclear();

  return 0;
}

module_init(init_lcd)
module_exit(exit_lcd)
